import pathlib
from typing import Union
import click
import pandas as pd


def has_valid_chromsizes(chrom_sizes: Union[str, pathlib.Path]) -> bool:
    """Check if the chromosome sizes file is valid"""

    if not chrom_sizes:
        return False

    if not pathlib.Path(chrom_sizes).is_file():
        return False

    return True


def has_tracks_to_convert(df: pd.DataFrame) -> bool:
    """Check if the design has tracks that need to be converted"""

    bed_files = df[df["ext"] == "bed"]
    if bed_files.empty:
        return False

    return True



import os
import click
from typing import Literal, Union

class OptionEatAll(click.Option):

    def __init__(self, *args, **kwargs):
        self.save_other_options = kwargs.pop('save_other_options', True)
        nargs = kwargs.pop('nargs', -1)
        assert nargs == -1, 'nargs, if set, must be -1 not {}'.format(nargs)
        super(OptionEatAll, self).__init__(*args, **kwargs)
        self._previous_parser_process = None
        self._eat_all_parser = None

    def add_to_parser(self, parser, ctx):

        def parser_process(value, state):
            # method to hook to the parser.process
            done = False
            value = [value]
            if self.save_other_options:
                # grab everything up to the next option
                while state.rargs and not done:
                    for prefix in self._eat_all_parser.prefixes:
                        if state.rargs[0].startswith(prefix):
                            done = True
                    if not done:
                        value.append(state.rargs.pop(0))
            else:
                # grab everything remaining
                value += state.rargs
                state.rargs[:] = []
            value = tuple(value)

            # call the actual process
            self._previous_parser_process(value, state)

        retval = super(OptionEatAll, self).add_to_parser(parser, ctx)
        for name in self.opts:
            our_parser = parser._long_opt.get(name) or parser._short_opt.get(name)
            if our_parser:
                self._eat_all_parser = our_parser
                self._previous_parser_process = our_parser.process
                our_parser.process = parser_process
                break
        return retval


@click.group()
def cli():
    """Tracknado: A tool for generating UCSC track hubs
    Tracknado go BRRRR!
    """

@cli.command()
@click.option("-i", "--input-files", help="Input files", cls=OptionEatAll, type=list)
@click.option("-o", "--output", help="Design name", required=True)
@click.option(
    "--preset",
    help="Adjust design for specific use cases. Choose from: seqnado",
    type=click.Choice(["seqnado"]),
    default=None,
)

def design(input_files: list, output: str, preset: Union[None, Literal["seqnado"]]):

    from tracknado.api import TrackFiles

    tf = TrackFiles(input_files, infer_subgroups=True, infer_attributes=True)

    if preset == "seqnado":
        tf.files = tf.files.assign(experiment=lambda df: df["fn"].apply(lambda x: x.parent.parent.parent.name))

    tf.files.to_csv(output)



@cli.command()
@click.option("-i", "--input-files", help="Input files", cls=OptionEatAll, type=tuple)
@click.option("-o", "--output", help="Hub output directory", required=True)
@click.option("--save-hub-design", help="Save hub design to file", type=str)
@click.option(
    "-d",
    "--details",
    help="Extra details for each file. MUST contain: 'fn' column",
)
@click.option(
    "--infer-details",
    help="Infer details from filenames. Requires SAMPLENAME_[ANTIBODY]_[REPLICATE]",
    is_flag=True,
)
@click.option(
    "--hub-name",
    help="Name of hub to generate",
    default="HUB",
    required=True,
)
@click.option(
    "--hub-email",
    help="Email address for hub",
    default="alastair.smith@ndcls.ox.ac.uk",
    required=True,
)
@click.option(
    "--genome-name",
    help="Name of genome",
    default="hg19",
    required=True,
)
@click.option(
    "--custom-genome",
    help="Determines if this is a custom genome",
    is_flag=True,
    default=False,
)
@click.option(
    "--genome-twobit",
    help="Twobit file required for custom genome",
)
@click.option(
    "--genome-organism",
    help="Organism name required for custom genome",
    default="NA",
)
@click.option(
    "--color-by",
    help="Name of column in details dataframe to use for track colors",
    multiple=True,
    default=None,
)
@click.option("--description-html", help="Path to HTML file with hub description")
@click.option(
    "--supergroup-by",
    help="Column(s) defining supertracks",
    multiple=True,
    default=None,
)
@click.option(
    "--subgroup-by",
    help="Column(s) defining subgroupings for composite tracks. If any value is provided a composite track will be generated for each filetype",
    multiple=True,
    default=None,
)
@click.option(
    "--composite-by",
    help="Column(s) defining composite tracks.",
    multiple=True,
    default=None,
)
@click.option(
    "--overlay-by",
    help="Column(s) defining overlay tracks",
    multiple=True,
    default=None,
)
@click.option(
    "--convert-files",
    help="Convert files to UCSC compatible format",
    is_flag=True,
    default=False,
)
@click.option(
    "--chrom-sizes",
    help="Path to chrom.sizes file",
    default=None,
)
def create(*args, **kwargs):

    """Create a UCSC track hub from a set of files

    The options provided will determine how the files are grouped into tracks and
    supertracks. 

    There are two ways to provide the files to be included in the hub. The first is to provide
    a list of files using the --input-files option. The second is to provide a design file using
    the --details option.

    File attributes can be inferred from the filenames using the --infer-details option. This
    option requires that the filenames are in the format SAMPLENAME_[ANTIBODY]_[REPLICATE].
    The attributes will be inferred from the filename and the following columns will be added
    to the design file (if they are present): 'sample', 'antibody', 'replicate'.
    
    The --supergroup-by, --subgroup-by, --composite-by, and --overlay-by options can be used to specify
    columns in the design file that will be used to group the files into tracks. The columns can be
    specified multiple times to group by multiple columns. For example, if the design file contains
    columns 'sample', 'antibody', and 'replicate' then the following options will group the files
    into supertracks by sample, then subtracks by antibody, then composite tracks by replicate:

    --supergroup-by sample --subgroup-by antibody --composite-by replicate

    The --color-by option can be used to specify a column in the design file that will be used to
    assign colors to the tracks. The colors will be assigned in the order that the values appear in
    the column unless colors are specified in the design file using the 'color' column.

    The --convert-files option can be used to convert the input files to UCSC compatible format. This
    option is only available for bed files at this time. The files will be converted in the current
    working directory and the converted files will be used to generate the hub. The original files
    will be left in place.

    A custom genome can be specified using the --custom-genome option. This option requires that
    the --genome-twobit and --genome-organism options are also provided.

    The --description-html option can be used to provide a path to an HTML file that will be used
    as the hub description. The HTML file will be copied to the hub directory and the path to the
    file will be added to the hub.txt file.

    """

    # sys.tracebacklimit = 0
    import pandas as pd
    from tracknado.api import TrackDesign, HubGenerator

    # Fix cli to api differences
    kwargs["color_by"] = kwargs["color_by"][0] if kwargs["color_by"] else None


    # Get design
    if not kwargs["details"]:
        assert kwargs[
            "input_files"
        ], "Input file MUST be provided if not specifying by a design file"

        kwargs.pop("details")

        design = TrackDesign.from_files(
            kwargs["input_files"],
            infer_attributes=kwargs["infer_details"],
            chromosome_sizes=kwargs["chrom_sizes"],
            **kwargs
        )

    else:
        details = pd.read_csv(kwargs["details"], sep=r"\s+|\t|,", engine="python")
        kwargs.pop("input_files")
        kwargs.pop("details")

        design = TrackDesign.from_design(
            details,
            infer_attributes=kwargs["infer_details"],
            chromosome_sizes=kwargs["chrom_sizes"],
            **kwargs
        )


    hub = HubGenerator(
        outdir=kwargs["output"],
        track_design=design,
        genome=kwargs["genome_name"],
        hub_name=kwargs["hub_name"],
        hub_email=kwargs["hub_email"],
        custom_genome=kwargs["custom_genome"],
        genome_twobit=kwargs["genome_twobit"],
        genome_organism=kwargs["genome_organism"],
        description_html=kwargs["description_html"],
    )

    hub.stage_hub()

    if kwargs["save_hub_design"]:
        design.to_pickle(kwargs["save_hub_design"])
    



@cli.command()
@click.option("-i", "--input-files", help="Input files", cls=OptionEatAll, type=tuple)
@click.option("-o", "--output", help="Hub output directory", required=True)
@click.option(
    "--hub-name",
    help="Name of hub to generate",
    default="HUB",
)
@click.option(
    "--hub-email",
    help="Email address for hub",
    default="alastair.smith@ndcls.ox.ac.uk",
)
@click.option(
    "--genome-name",
    help="Name of genome",
    default="hg19",
)
@click.option(
    "--custom-genome",
    help="Determines if this is a custom genome",
    is_flag=True,
    default=False,
)
@click.option(
    "--genome-twobit",
    help="Twobit file required for custom genome",
)
@click.option(
    "--genome-organism",
    help="Organism name required for custom genome",
    default="NA",
)
@click.option(
    "--color-by",
    help="Name of column in details dataframe to use for track colors",
    multiple=True,
    default=None,
    type=list
)
@click.option("--description-html", help="Path to HTML file with hub description")
@click.option(
    "--supergroup-by",
    help="Column(s) defining supertracks",
    multiple=True,
    default=None,
)
@click.option(
    "--subgroup-by",
    help="Column(s) defining subgroupings for composite tracks. If any value is provided a composite track will be generated for each filetype",
    multiple=True,
    default=None,
)
@click.option(
    "--composite-by",
    help="Column(s) defining composite tracks.",
    multiple=True,
    default=None,
)
@click.option(
    "--overlay-by",
    help="Column(s) defining overlay tracks",
    multiple=True,
    default=None,
)
def merge(*args, **kwargs):
    """Merge multiple seqnado created hubs into a single hub.
    
    This command will merge multiple seqnado created hubs into a single hub. 
    
    This requires that the hubs were created with the tracknado create command and that the hub designs were saved using the --save-hub-design option.
    """

    assert all(os.path.exists(hub) for hub in kwargs["input_files"]), "All input files must exist"

    import pandas as pd
    from tracknado.api import TrackDesign, HubGenerator

    for ii, hub in enumerate(kwargs["input_files"]):
        if ii == 0:
            design = TrackDesign.from_pickle(hub)
        else:
            design = design + TrackDesign.from_pickle(hub) # type: ignore
        

    hub = HubGenerator(
        outdir=kwargs["output"],
        track_design=design, # type: ignore
        genome=kwargs["genome_name"],
        hub_name=kwargs["hub_name"],
        hub_email=kwargs["hub_email"],
        custom_genome=kwargs["custom_genome"],
        genome_twobit=kwargs["genome_twobit"],
        genome_organism=kwargs["genome_organism"],
        description_html=kwargs["description_html"],
    )

    hub.stage_hub()











import os
import pathlib
import shutil
import re
from typing import Dict, Tuple, Union, List

import tempfile
import numpy as np
import pandas as pd
import seaborn as sns
import trackhub
import subprocess

from .track import get_file_attributes, get_groups_from_design_matrix
from .hub_setup import get_genome_file, make_track_palette
from .grouping import (
    get_subgroup_definitions,
    add_composite_tracks_to_container,
    add_overlay_track_to_container,
    add_hub_group,
)


def stage_hub(
    hub: trackhub.Hub,
    genome: trackhub.Genome,
    outdir: str,
    description_url_path: str = None,
):

    with tempfile.TemporaryDirectory() as tmpdir:
        trackhub.upload.stage_hub(hub, staging=tmpdir)

        if description_url_path:
            description_basename = os.path.basename(description_url_path)
            with open(os.path.join(tmpdir, f"{hub.hub}.hub.txt"), "a") as hubtxt:
                hubtxt.write("\n")
                hubtxt.write(f"descriptionUrl {genome.genome}/{description_basename}\n")

            shutil.copy(
                description_url_path,
                os.path.join(tmpdir, genome.genome),
            )

        # Copy to the new location
        shutil.copytree(
            tmpdir,
            outdir,
            dirs_exist_ok=True,
            symlinks=False,
        )

        subprocess.run(["chmod", "-R", "2755", outdir])
    

def get_grouping_columns(cols):
    try:
        if len(cols) > 1:
            return cols
        else:
            return cols[0]
    except IndexError:
        return None


def make_hub(
    files: Tuple,
    output: str,
    details: Union[str, pd.DataFrame] = None,
    group_composite: Tuple[str] = None,
    group_overlay: Tuple[str] = None,
    **kwargs,
):

    # Set-up varibles
    group_composite = get_grouping_columns(group_composite)
    group_overlay = get_grouping_columns(group_overlay)

    # Get file attributes
    df_file_attributes = get_file_attributes(
        files,
        convert=kwargs.get("convert", False),
        chrom_sizes=kwargs.get("chrom_sizes", None),
    )

    # Design matrix in the format: filename samplename ATTRIBUTE_1 ATTRIBUTE_2 ...
    if isinstance(details, str):
        df_details = pd.read_csv(
            details, sep=r"\s+|\t|,", engine="python", index_col="filename"
        )
    elif isinstance(details, pd.DataFrame):
        df_details = details
    else:
        df_details = pd.DataFrame()
        # raise RuntimeError("File detail not provided")

    if not df_details.empty:
        df_file_attributes = get_groups_from_design_matrix(
            df_file_attributes=df_file_attributes, df_design=df_details
        )

    # Create hub
    hub = trackhub.Hub(
        hub=kwargs["hub_name"],
        short_label=kwargs.get("hub_short_label", kwargs["hub_name"]),
        long_label=kwargs.get("hub_long_label", kwargs["hub_name"]),
        email=kwargs["hub_email"],
    )

    # Genome
    custom_genome = kwargs.get("custom_genome", False)
    genome = get_genome_file(**kwargs)

    # Create genomes file
    genomes_file = trackhub.GenomesFile()

    # Create trackdb
    trackdb = trackhub.TrackDb()

    # Add these to the hub
    hub.add_genomes_file(genomes_file)
    genome.add_trackdb(trackdb)
    genomes_file.add_genome(genome)

    # # Create composite track definitions
    subgroup_definitions = get_subgroup_definitions(
        df_file_attributes=df_file_attributes, grouping_columns=group_composite
    )

    color_tracks_by = list(kwargs.get("color_by", ("name",)))

    color_mapping = make_track_palette(
        df_file_attributes=df_file_attributes,
        palette=kwargs.get("palette", "hls"),
        color_by=color_tracks_by,
    )

    # Make supertracks to hold any groupings
    supertracks = dict()

    # Composite tracks
    if group_composite:
        for group_name, df in df_file_attributes.groupby(group_composite):
            supertracks[group_name] = trackhub.SuperTrack(name=group_name)

            add_composite_tracks_to_container(
                container=supertracks[group_name],
                track_details=df,
                subgroup_definitions=subgroup_definitions,
                color_by=color_tracks_by,
                color_mapping=color_mapping,
                track_suffix=group_name,
                custom_genome=custom_genome,
                hub=hub,
                genome_twobit=kwargs["genome_twobit"],
            )
            add_hub_group(
                container=supertracks[group_name], hub=hub, custom_genome=custom_genome
            )
    else:
        add_composite_tracks_to_container(
            container=trackdb,
            track_details=df_file_attributes,
            subgroup_definitions=subgroup_definitions,
            color_by=color_tracks_by,
            color_mapping=color_mapping,
            custom_genome=custom_genome,
            hub=hub,
            genome_twobit=kwargs["genome_twobit"],
        )

    if group_overlay:
        for group_name, df in df_file_attributes.groupby(group_overlay):

            supertrack_exists = False
            if not group_name in supertracks:
                supertrack_exists = True
                supertracks[group_name] = trackhub.SuperTrack(name=group_name)

            add_overlay_track_to_container(
                track_name=group_name,
                container=supertracks[group_name],
                track_details=df,
                color_by=color_tracks_by,
                color_mapping=color_mapping,
                custom_genome=custom_genome,
                hub=hub,
            )

            if not supertrack_exists:
                add_hub_group(
                    container=supertracks[group_name],
                    hub=hub,
                    custom_genome=custom_genome,
                )

    trackdb.add_tracks(supertracks.values())

    ## Add generic tracks
    for track in kwargs["generic_tracks"]:
        track_name = trackhub.helpers.sanitize(os.path.basename(t))
        track_path = os.path.abspath(track)
        t = trackhub.Track(
            name=track_name,
            source=track_path,
            tracktype=os.path.basename(track_name).split(".")[1],
        )

        trackdb.add_tracks(track)

    # Copy hub to the correct directory
    stage_hub(
        hub=hub,
        genome=genome,
        outdir=output,
        description_url_path=kwargs.get("description_html"),
    )



import pandas as pd
import pathlib
from typing import Tuple
import seaborn as sns
import trackhub
from typing import Dict
import subprocess
import numpy as np
from collections import defaultdict


def convert_to_bigbed(bed: str, chrom_sizes: str, out: str):

    cmd = f"""
            export LC_ALL=C &&
            sort -k1,1 -k2,2n {bed} | cut -f 1-3 | grep -P 'chr(\d+|X|Y)' > {bed}.tmp && 
            bedToBigBed {bed}.tmp {chrom_sizes} {out}
            rm {bed}.tmp
           """
    subprocess.run(cmd, shell=True)
    return out



def get_file_attributes(files: Tuple, convert: bool = False, chrom_sizes: str = ""):

    if convert:
        files_new = []
        for fn in files:
            if fn.endswith(".bed"):
                files_new.append(convert_to_bigbed(fn, chrom_sizes, fn.replace(".bed", ".bigBed")))
            else:
                files_new.append(fn)
        files = tuple(files_new)
    
    df = pd.Series(files).to_frame("fn")
    paths = [pathlib.Path(fn) for fn in df["fn"].values]
    df["path"] = [str(p.absolute().resolve()) for p in paths]
    df["basename"] = [p.name for p in paths]
    df["name"] = [p.stem for p in paths]
    df["ext"] = [p.suffix.strip(".") for p in paths]

    # Deal with duplicate names
    basenames = defaultdict(int)
    for row in df.itertuples():
        basenames[row.basename] += 1
        if basenames[row.basename] > 1:

            name = f"{row.name}_{basenames[row.basename]}"
            basename = f"{row.basename}_{basenames[row.basename]}.{row.ext}"
            df.loc[row.Index, "name"] = name
            df.loc[row.Index, "basename"] = basename


    return df


def get_groups_from_regex(df_file_attributes: pd.DataFrame, pattern: str):

    groups = df_file_attributes["basename"].str.extract(pattern).reset_index()
    return pd.concat([df_file_attributes, groups], axis=1)


def get_groups_from_design_matrix(
    df_file_attributes: pd.DataFrame, df_design: pd.DataFrame
):
    return (
        df_file_attributes.set_index("fn")
        .join(df_design)
        .reset_index()
        .rename(columns={"index": "fn"})
    )

def get_track_subgroups(
    track_details: tuple,
    subgroup_definitions: Dict[str, trackhub.SubGroupDefinition],
):
    return "_".join(
        [str(getattr(track_details, subgroup)).lower() for subgroup in subgroup_definitions]
    )


import click
from .make_hub import get_file_attributes, make_hub
import pandas as pd


def categorise_capcruncher_tracks(df_file_attributes: pd.DataFrame):
    mapping = {
        "raw": "Replicates",
        "normalised": "ReplicatesScaled",
        "summary": "SamplesSummarised",
        "subtraction": "SamplesCompared",
    }

    return [
        mapping[k]
        for _, method in df_file_attributes["method"].iteritems()
        for k in mapping
        if k in method
    ]


def get_ngs_pipeline_attributes(df_file_attributes: pd.DataFrame):

    chipseq_pattern = r"(?P<samplename>.*?)_(?P<antibody>.*)_(?:.*).(bigWig|bed|bigBed)"
    generic_pattern = r"(?P<samplename>.*?).(bigWig|bed|bigBed)"
    attributes = df_file_attributes["basename"].str.extract(chipseq_pattern)

    if attributes["samplename"].dropna().shape[0]:
        attributes = df_file_attributes["basename"].str.extract(
            generic_pattern)

    return attributes


@click.command()
@click.argument("files", nargs=-1)
@click.option("-o", "--output", help="Output hub directory name", default="HUB/")
@click.option(
    "-d",
    "--details",
    help="Extra details for each file. MUST contain: filename and samplename columns",
)
@click.option("-g", "--group-by", help="Column in the provided details to group tracks")
@click.option("--description-html", help="Path to HTML file with hub description")
@click.option(
    "-m",
    "--mode",
    help="Presets for CapCruncher and NGS-Pipeline files",
    type=click.Choice(["capcruncher", "ngs-pipeline", "custom"]),
    default="custom",
)
@click.option(
    "-t",
    "--generic-tracks",
    help="Ungrouped tracks to add to the hub",
    multiple=True,
)
@click.option(
    "--hub-name",
    help="Name of hub to generate",
    default="HUB",
)
@click.option(
    "--hub-email",
    help="Email address for hub",
    default="alastair.smith@ndcls.ox.ac.uk",
)
@click.option(
    "--genome-name",
    help="Name of genome",
    default="hg19",
)
@click.option(
    "--custom-genome",
    help="Determines if this is a custom genome",
    is_flag=True,
    default=False,
)
@click.option(
    "--genome-twobit",
    help="Twobit file required for custom genome",
)
@click.option(
    "--genome-organism",
    help="Organism name required for custom genome",
    default="NA",
)
@click.option(
    "--color-by",
    help="Name of column in details dataframe to use for track colors",
    multiple=True,
    default=("name", ),
)
@click.option(
    "--group-overlay",
    help="Attributes to use for a grouped overlay track",
    multiple=True,
    default=None,
)
@click.option(
    "--group-composite",
    help="Attributes to use for a grouped composite track",
    multiple=True,
    default=None,
)
@click.option(
    "--convert",
    help="Convert files to UCSC compatible format",
    is_flag=True,
    default=False,
)
@click.option(
    "--chrom-sizes",
    help="Path to chrom.sizes file",
    default=None,
)
def create_hub(files, **kwargs):

    if kwargs["mode"] == "capcruncher":

        df = get_file_attributes(files)
        attributes = df["basename"].str.extract(
            r"(?P<samplename>.*?)\.(?P<method>.*?)\.(?P<viewpoint>.*?)\.(?P<file_type>.*)"
        )

        df_details = df.join(attributes)
        df_details["track_category"] = categorise_capcruncher_tracks(
            df_details)
        df_details = (df_details[["fn", "samplename","viewpoint", "track_category"]]
                                .set_index("fn"))

        make_hub(
            files=files,
            details=df_details,
            group_by="track_category",
            **{k: v for k, v in kwargs.items() if not k in ["files", "details", "group_by"]}
        )

    elif not kwargs["details"]:
         df = get_file_attributes(files)
         make_hub(
            files=files,
            **{kw:val for kw, val in kwargs.items() if not "details" in kw}
            )

    else:
        make_hub(files, **kwargs)




import os
import pathlib
import pickle
import re
import shutil
import subprocess
import tempfile
from collections import defaultdict, namedtuple
from typing import Dict, List, Literal, Tuple, Union
import hashlib
import json
from loguru import logger

import pandas as pd
import seaborn as sns
import trackhub

from tracknado.utils import has_valid_chromsizes, has_tracks_to_convert


def fix_duplicate_names(df: pd.DataFrame):
    duplicate_counts = defaultdict(int)

    for row in df.itertuples():
        duplicate_counts[row.basename] += 1
        if duplicate_counts[row.basename] > 1:
            name = f"{row.name}_{duplicate_counts[row.basename]}"
            basename = f"{row.basename}_{duplicate_counts[row.basename]}.{row.ext}"
            df.loc[row.Index, "name"] = name
            df.loc[row.Index, "basename"] = basename

def get_hash(obj: object) -> str:
    """Get hash of object"""
    return hashlib.md5(json.dumps(obj).encode("utf-8")).hexdigest()

def get_hash_for_df(df: pd.DataFrame, columns: Union[List[str], pd.Index] = None) -> List[str]:

    if columns is None:
        columns = df.columns
    
    hashes = []
    for row in df.itertuples():
        x = tuple([getattr(row, x) for x in columns])
        hash = get_hash(x)
        hashes.append(hash)
    
    return hashes

class TrackFiles:
    def __init__(
        self,
        files: Union[List[str], List[pathlib.Path], pd.DataFrame],
        infer_subgroups: bool = False,
        infer_attributes: bool = False,
        deduplicate: bool = False,
        convert_files: bool = False,
        chromosome_sizes: Union[pathlib.Path, str] = "",
        **kwargs,
    ) -> None:

        self.files = self.get_file_attributes(files)  # type: ignore

        if (self.files["ext"] == "bed").any() and not convert_files:
            raise ValueError(
                "BED files detected. Please set convert_files=True to convert to BigBed"
            )
        else:
            self.convert_tracks_to_ucsc_format(chrom_sizes=chromosome_sizes, outdir=".")

        if deduplicate:
            self.fix_duplicate_names()

        if infer_attributes:
            self.files = self.infer_attributes_from_file_names()

        self.subgroup_columns = (
            self.infer_subgroup_columns() if infer_subgroups else None
        )

    def get_file_attributes(
        self, files: Union[pd.DataFrame, List[Union[str, pathlib.Path]]]
    ) -> pd.DataFrame:

        if isinstance(files, pd.DataFrame):
            assert "fn" in files.columns, "DataFrame must have a column named 'fn'"
            df = files
        else:
            df = pd.Series(files).to_frame("fn")

        paths = [pathlib.Path(fn) for fn in df["fn"].values]
        df["path"] = [str(p.absolute().resolve()) for p in paths]
        df["basename"] = [p.name for p in paths]
        df["name"] = [p.stem for p in paths]
        df["ext"] = [p.suffix.strip(".") for p in paths]
        return df

    def infer_subgroup_columns(self) -> List[str]:
        return self.files.columns.difference(
            ["fn", "path", "basename", "name", "ext"]
        ).tolist()

    def fix_duplicate_names(self):
        duplicate_counts = defaultdict(int)

        for row in self.files.itertuples():
            duplicate_counts[row.basename] += 1
            if duplicate_counts[row.basename] > 1:
                name = f"{row.name}_{duplicate_counts[row.basename]}"
                basename = f"{row.basename}_{duplicate_counts[row.basename]}.{row.ext}"
                self.files.loc[row.Index, "name"] = name
                self.files.loc[row.Index, "basename"] = basename

    def convert_tracks_to_ucsc_format(
        self, chrom_sizes: Union[str, pathlib.Path], outdir: Union[str, pathlib.Path]
    ) -> None:
        """Convert tracks to UCSC format"""


        if has_tracks_to_convert(self.files):


            if not has_valid_chromsizes(chrom_sizes):
                raise ValueError(
                    f"Chromosome sizes file {chrom_sizes} does not exist")

            outdir = pathlib.Path(outdir)
            outdir.mkdir(exist_ok=True)

            # convert bed to bigBed
            bed_files = self.files[self.files["ext"] == "bed"]
            for bed_file in bed_files.itertuples():

                print(f"Converting {bed_file.name} to BigBed format")

                bed_file_path = pathlib.Path(bed_file.path).absolute()
                bed_file_basename = bed_file_path.name.removesuffix(".bed")


                cmd_sort = [
                    "sort",
                    "-k1,1",
                    "-k2,2n",
                    str(bed_file_path),
                    "-o",
                    bed_file_path.with_suffix(".sorted.bed"),
                ]

                cmd_convert = [
                    "bedToBigBed",
                    str(bed_file_path.with_suffix(".sorted.bed")),
                    str(chrom_sizes),
                    f"{outdir.joinpath(bed_file_basename)}.bigBed",
                ]

                subprocess.run(cmd_sort, check=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
                subprocess.run(cmd_convert, check=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)

                # Remove the sorted bed file
                bed_file_path.with_suffix(".sorted.bed").unlink()

                # Update the file name in the dataframe
                self.files.loc[bed_file.Index, "path"] = outdir.joinpath(bed_file_basename).with_suffix(".bigBed")
                self.files.loc[bed_file.Index, "ext"] =  "bigBed"
                self.files.loc[bed_file.Index, "fn"] = self.files.loc[bed_file.Index, "path"]        


    def infer_attributes_from_file_names(
        self,
        regex: str = r"(?P<sample_name>.*?)(?:_(?P<antibody>.*?))?(?:_(?P<replicate>[0-9]+))?$",
    ) -> pd.DataFrame:
        """Infer attributes from file names using a regex"""

        df = self.files.copy()
        df_attributes = df["name"].str.extract(regex)

        # Remove any columns that are all NaN
        df_attributes = df_attributes.loc[:, df_attributes.notnull().any()]

        return df.join(df_attributes)

class TrackDesign:
    def __init__(
        self,
        details: pd.DataFrame,
        color_by: List[str] = None,
        subgroup_by: List[str] = None,
        overlay_by: List[str] = None,
        supergroup_by: List[str] = None,
        **kwargs,
    ):

        self.details = details
        self._supertrack_columns = list(supergroup_by) if supergroup_by else list()
        self._overlay_columns = list(overlay_by) if overlay_by else list()
        self._subgroup_columns = list(subgroup_by) if subgroup_by else list()
        self.subgroup_definitions = list() if subgroup_by else None
        self._color_columns = list(color_by) if color_by else list()

        self._add_subgroupings(supergroup_by=self._supertrack_columns, subgroup_by=self._subgroup_columns)

        self.super_tracks = self._get_super_tracks()
        self._add_supertrack_indicators()

        self.composite_tracks = self._get_composite_tracks()
        self._add_composite_track_indicators()

        self.overlay_tracks = self._get_overlay_tracks()
        self._add_overlay_track_indicators()

        self._add_track_colors(color_by=color_by)

    @classmethod
    def from_files(cls, files: List[pathlib.Path], **kwargs) -> "TrackDesign":
        hub_files = TrackFiles(files, **kwargs)

        extra_kwargs = dict()
        if hub_files.subgroup_columns:
            extra_kwargs["subgroup_by"] = hub_files.subgroup_columns


        return cls(hub_files.files, **kwargs, **extra_kwargs)

    @classmethod
    def from_design(cls, design: pd.DataFrame, **kwargs) -> "TrackDesign":
        design = design.copy()
        design = TrackFiles(design, **kwargs).files
        return cls(design, **kwargs)
    
    @classmethod
    def from_pickle(cls, pickle_file: pathlib.Path) -> "TrackDesign":
        with open(pickle_file, "rb") as f:
            return pickle.load(f)

    def _add_track_colors(
        self,
        color_by: Union[str, List[str]] = None,
        pallet: str = "tab20",
        color_column: str = None,
    ) -> None:

        """Add a column to the details dataframe with a color for each track"""

        from PIL import ImageColor

        if color_by:
            if isinstance(color_by, str):
                color_by = [color_by]

            assert all([c in self.details.columns for c in color_by]), f"Color-By columns {color_by} missing"  # type: ignore

            try:
                # Get a pallette with enough colors for the unique groups in the details
                n_colors = len(self.details[color_by].drop_duplicates())
                colors = sns.color_palette(pallet, n_colors=n_colors).as_hex()  # type: ignore

                # Assign a color to each group
                color_dict = {}
                for i, group in enumerate(
                    self.details[color_by].drop_duplicates().itertuples()
                ):
                    color_dict[tuple([getattr(group, c) for c in color_by])] = colors[i]  # type: ignore

                # Add a column to the details dataframe with the color for each track
                self.details["color"] = self.details[color_by].apply(
                    lambda row: ImageColor.getrgb(color_dict[tuple([c for c in row])]),
                    axis=1,
                )

            except NameError:
                raise NameError(
                    "Pallette not found. Try one of the following: 'tab20', 'tab20b', 'tab20c'"
                )

        elif color_column:

            assert (
                color_column in self.details.columns
            ), f"Color column {color_column} missing"

            colors = []
            for i, color in enumerate(self.details[color_column]):
                if isinstance(color, tuple):
                    c = color
                elif isinstance(color, str):
                    if color.startswith("#"):
                        c = ImageColor.getrgb(color)
                    else:
                        c = color.split(",")
                        c = tuple([int(x) for x in c])
                else:
                    raise ValueError(
                        f"Color column {color_column} must be a tuple or string"
                    )

                colors.append(c)

            self.details["color"] = colors

    def _add_subgroup_definitions_to_df(
        self, df: pd.DataFrame, subgroup_by: List[str] = None
    ) -> pd.DataFrame:
        """Add a column to the details dataframe with a `trackhub.SubGroupDefinition` for each track"""

        assert all(
            [c in df.columns for c in subgroup_by]
        ), f"Subgroup-By columns {subgroup_by} missing"
        df = df.copy()

        # Loop through all columns provided and generate a subgroup definition for each
        subgroup_definitions = []
        for column in subgroup_by:
            # Get a list of unique values in the column
            unique_values = df[column].unique()
            subgroup_definition = trackhub.SubGroupDefinition(
                name=column,
                label=column,
                mapping={value: value for value in unique_values},
            )
            subgroup_definitions.append(subgroup_definition)

        # Add a column to the details dataframe with the subgroup definition for each track
        df["subgroup_names"] = [
            tuple([col for col in subgroup_by]) for i in range(df.shape[0])
        ]
        df["subgroup_definition"] = [subgroup_definitions for i in range(df.shape[0])]

        self.subgroup_definitions.extend(subgroup_definitions)

        return df

    def _add_subgroupings(
        self, supergroup_by: List[str] = None, subgroup_by: List[str] = None
    ) -> None:
        """Add a column to the details dataframe with a `trackhub.SubGroupDefinition` for each track.

        If `supergroup_by` is provided, the subgroup definitions will be added to the dataframe
        grouped by the supergroup columns.

        If `supergroup_by` is not provided, the subgroup definitions will be added to the dataframe
        as a single group.
        """

        if subgroup_by:

            assert all(
                [c in self.details.columns for c in subgroup_by]
            ), f"Subgroup-By columns {subgroup_by} missing"

            if supergroup_by:
                assert not any(
                    subgroup in supergroup_by for subgroup in subgroup_by
                ), f"SubGroup columns {subgroup_by} cannot be in SuperGroup columns {supergroup_by}"
          
                self.details = self.details.groupby(supergroup_by).apply(
                    self._add_subgroup_definitions_to_df, subgroup_by=subgroup_by
                )
            else:
                self.details = self._add_subgroup_definitions_to_df(
                    self.details, subgroup_by=subgroup_by
                )

    def _get_super_tracks(self) -> Dict[str, trackhub.SuperTrack]:
        """Generate a dictionary of SuperTracks from the details dataframe"""

        if self._supertrack_columns:
            assert all(
                [c in self.details.columns for c in self._supertrack_columns]
            ), f"SuperTrack columns {self._supertrack_columns} missing"

            supertracks = dict()
            for grouping, df in self.details.groupby(self._supertrack_columns):
                

                if isinstance(grouping, str):
                    track_id = (grouping,)
                elif len(grouping) == 1:
                    track_id = grouping
                else:
                    track_id = tuple(grouping)
                
                if len(track_id) == 1:
                    track_name = track_id[0]
                else:
                    track_name = "_".join(track_id)


                supertracks[get_hash(track_id)] = trackhub.SuperTrack(
                    name=track_name,
                )

        else:
            supertracks = dict()

        return supertracks

    def _add_supertrack_indicators(self):
        """Add a column to the details dataframe with a SuperTrack indicator for each track"""

        if self._supertrack_columns:
            assert all(
                [c in self.details.columns for c in self._supertrack_columns]
            ), f"SuperTrack columns {self._supertrack_columns} missing"

            self.details["supertrack"] = get_hash_for_df(self.details, self._supertrack_columns)
    
    def _get_composite_tracks(self) -> Dict[str, trackhub.CompositeTrack]:
        """Generate a dictionary of CompositeTracks from the details dataframe"""

        composite_tracks = dict()
        dimensions = dict(
                    zip(
                        [f"dim{d}" for d in ["X", "Y", "A", "B", "C", "D"]],
                        self._subgroup_columns,
                    )
                )
        
        if "supertrack" in self.details.columns:
            for (supertrack, ext) , df in self.details.groupby(["supertrack", "ext"]):
                

                supertrack_name = self.super_tracks[supertrack].name
                composite_name = "_".join([supertrack_name, ext])


                composite = trackhub.CompositeTrack(
                    name=composite_name,
                    tracktype=ext,
                    dimensions=" ".join([f"{k}={v}" for k, v in dimensions.items()])
                    if dimensions
                    else None,
                    sortOrder=" ".join([f"{k}=+" for k in self._subgroup_columns]),
                    visibility="hide",
                    dragAndDrop="subTracks",
                    allButtonPair="off",
                )

                composite.add_subgroups(self.subgroup_definitions)

                self.super_tracks[supertrack].add_tracks(composite)
                composite_tracks[get_hash((supertrack, ext))] = composite

        elif self._subgroup_columns:
            for ext, df in self.details.groupby("ext"):
                composite = trackhub.CompositeTrack(
                    name=ext,
                    tracktype=ext,
                    visibility="hide",
                    dragAndDrop="subTracks",
                    allButtonPair="off",
                    dimensions=" ".join([f"{k}={v}" for k, v in dimensions.items()])
                    if dimensions
                    else None,
                    sortOrder=" ".join([f"{k}=+" for k in self._subgroup_columns]),
                )

                composite.add_subgroups(self.subgroup_definitions)
                composite_tracks[get_hash((ext,))] = composite
        
        else:
            composite_tracks = dict()

        return composite_tracks

    def _add_composite_track_indicators(self):
        """Add a column to the details dataframe with a CompositeTrack indicator for each track"""

        if self.composite_tracks:
            composite_columns = ["supertrack"] if self._supertrack_columns else []
            composite_columns.append("ext")

            self.details["composite"] = get_hash_for_df(self.details, composite_columns)

            assert self.details["composite"].isin(self.composite_tracks.keys()).all(), (
                "Composite tracks not found in details dataframe"
            )

    def _get_overlay_tracks(self):
        """Generate a dictionary of OverlayTracks from the details dataframe"""

        if self._overlay_columns:
            assert all(
                [c in self.details.columns for c in self._overlay_columns]
            ), f"Overlay columns {self._overlay_columns} missing"

            overlay_tracks = dict()
            overlay_columns = list(self._overlay_columns) if not isinstance(self._overlay_columns, str) else [self._overlay_columns,]

            if "supertrack" in self.details.columns:

                for (supertrack, overlay) , df in self.details.groupby(
                    ["supertrack", *self._overlay_columns]
                ):
                    
                    supertrack_name = self.super_tracks[supertrack].name
                    overlay_name = "_".join([supertrack_name, *overlay]) + "_overlay"

                    overlay_track = trackhub.AggregateTrack(
                        aggregate="transparentOverlay",
                        name=overlay_name,
                    )

                    self.super_tracks[supertrack].add_tracks(overlay)
                    overlay_tracks[get_hash(tuple([supertrack, overlay]))] = overlay_track

            else:
                for overlay, df in self.details.groupby(self._overlay_columns):

                    overlay_name = "_".join(overlay) if isinstance(overlay, tuple) else overlay
                    overlay_id = tuple(overlay) if isinstance(overlay, tuple) else (overlay,)

                    overlay_track = trackhub.AggregateTrack(
                        aggregate="transparentOverlay",
                        name=overlay_name,
                    )
                    overlay_tracks[get_hash(overlay_id)] = overlay_track

        else:
            overlay_tracks = dict()

        return overlay_tracks

    def _add_overlay_track_indicators(self):
        """Add a column to the details dataframe with an OverlayTrack indicator for each track"""

        if self._overlay_columns:

            overlay_columns = (
                ["supertrack"] if self._supertrack_columns else []
            )
            overlay_columns.extend(self._overlay_columns)

            self.details["overlay"] = get_hash_for_df(self.details, overlay_columns)

            assert self.details["overlay"].isin(self.overlay_tracks.keys()).all(), (
                "Overlay tracks not found in details dataframe"
            )


    def __add__(self, other):
        """Combine two HubDesign objects"""
            
        assert isinstance(other, TrackDesign), "Can only combine HubDesign objects"

        self.details = pd.concat([self.details, other.details], ignore_index=True)
        fix_duplicate_names(self.details)

        if any([self._supertrack_columns, other._supertrack_columns]):
            self._supertrack_columns = sorted(
                list(set(self._supertrack_columns + other._supertrack_columns))
            )
            self._add_supertrack_indicators()


        if any([self._subgroup_columns, other._subgroup_columns]):
            self._subgroup_columns = sorted(
                list(set(self._subgroup_columns + other._subgroup_columns))
            )
            self._add_composite_track_indicators()

        
        if any([self._overlay_columns, other._overlay_columns]):
            self._overlay_columns = sorted(
                list(set(self._overlay_columns + other._overlay_columns))
            )
            self._add_overlay_track_indicators()

        return self

    def to_pickle(self, path: str) -> None:
        """Save the design to a pickle file"""
        with open(path, "wb") as f:
            pickle.dump(self, f)


class HubGenerator:
    def __init__(
        self,
        hub_name: str,
        genome: str,
        track_design: TrackDesign,
        outdir: pathlib.Path,
        description_html: pathlib.Path = None,
        hub_email: str = "",
        custom_genome: bool = False,
        genome_twobit: pathlib.Path = None,
        genome_organism: str = None,
        genome_default_position: str = "chr1:10000-20000",
    ):

        # Basic parameters for hub creation
        self.hub_name = hub_name
        self.genome_name = genome
        self.track_design = track_design
        self.outdir = outdir
        self.custom_genome = custom_genome
        self.description_url_path = description_html

        # Parameters for custom genomes
        self._genome_twobit = genome_twobit
        self._genome_organism = genome_organism
        self._genome_default_position = genome_default_position

        # Create the basic hub
        self._hub = trackhub.Hub(
            hub_name, short_label=hub_name, long_label=hub_name, email=hub_email
        )
        
        self.trackdb = trackhub.TrackDb()
        _genome = self._get_genome_file()  # type: ignore
        _genomes_file = trackhub.GenomesFile()

        # Add these to the hub
        _genome.add_trackdb(self.trackdb)
        self._hub.add_genomes_file(_genomes_file)
        _genomes_file.add_genome(_genome)

        self._add_tracks_to_hub()

    def _add_tracks_to_hub(self) -> None:
        # Loop through each entry in the details dataframe
        
        for row in self.track_design.details.itertuples():

            has_composite = False
            has_overlay = False

            # If the row has a "composite" attribute
            if hasattr(row, "composite"):
                has_composite = True
                composite_track = self.track_design.composite_tracks[row.composite]
                # Create a new track and add it as a subtrack to the composite track
                track = self._get_track(row, suffix=f"_{composite_track.name}")
                composite_track.add_subtrack(track)

            # If the row has an "overlay" attribute
            if hasattr(row, "overlay"):
                has_overlay = True
                overlay_track = self.track_design.overlay_tracks[row.overlay]
                # Create a new track and add it to the overlay track
                track = self._get_track(row, suffix=f"_{overlay_track.name}")

                # Ignore the track if it is not a signal track e.g. bigWig
                if track.tracktype not in ["bigWig", ]:
                    logger.warning(f"Track {track.name} is not a signal track and will be ignored for the overlay track {overlay_track.name}")
                else:
                    overlay_track.add_subtrack(track)

            # If the row doesn't have a "supertrack" attribute
            if not hasattr(row, "supertrack") and not has_composite and not has_overlay:
                # Create a new track and add it to the trackdb
                track = self._get_track(row)
                self.trackdb.add_tracks(track)

        # Add the supertracks or composite/overlay tracks to the trackdb
        if self.track_design.super_tracks:
            tracks = self.track_design.super_tracks.values()
            
            # Ensure the composite and/or overlay tracks have the group attribute set
            if self.custom_genome:
                for t in [*self.track_design.composite_tracks.values(), *self.track_design.overlay_tracks.values()]:
                    t.add_params(group=self._hub.hub)

        else:
            tracks = [*self.track_design.composite_tracks.values(), *self.track_design.overlay_tracks.values()]

        # Add the composite/overlay and supertracks to the trackdb
        for ii, track in enumerate(tracks):
            # Add group if custom genome
            if self.custom_genome:
                track.add_params(group=self._hub.hub)
            self.trackdb.add_tracks(track)

    

    def _get_track(self, track: namedtuple, suffix: str = "") -> trackhub.Track:
        """Generate a trackhub.Track object from a row in the details dataframe"""


        extra_kwargs = dict()
        if hasattr(track, "color"):
            extra_kwargs["color"] = ",".join([str(x) for x in track.color])
        
        if hasattr(track, "subgroup_names"):
            extra_kwargs["subgroups"] = {subgroup_name: getattr(track, subgroup_name) 
                                         for subgroup_name in track.subgroup_names}

        if self.custom_genome:
            extra_kwargs["group"] = self._hub.hub

        if track.ext == "bigWig":
            extra_kwargs.update(
                {
                    "maxHeightPixels": "100:50:11",
                    "visibility": "full",
                    "viewLimits": "0:100",
                    "autoScale": "on",
                    "windowingFunction": "mean",
                }
            )
        
        elif track.ext == "bigBed":
            extra_kwargs.update(
                {
                    "visibility": "pack",
                }
            )
        
        return  trackhub.Track(
                name="".join([trackhub.helpers.sanitize(track.name), suffix]),
                shortLabel=" ".join(re.split(r"[.|_|\s+|-]", track.name)),
                longLabel=" ".join(re.split(r"[.|_|\s+|-]", track.name)),
                source=str(track.path),
                tracktype=track.ext,
                **extra_kwargs,
               
            )
      
    def _get_genome_file(self) -> trackhub.Genome:

        if not self.custom_genome:
            genome = trackhub.Genome(self.genome_name)
            groups_file = None
        else:
            genome = trackhub.Assembly(
                genome=self.genome_name,
                twobit_file=self._genome_twobit,
                organism=self._genome_organism,
                defaultPos=self._genome_default_position,
            )

            groups_file = trackhub.GroupsFile(
                [
                    trackhub.GroupDefinition(
                        name=self.hub_name, priority=1, default_is_closed=False
                    ),
                ]
            )

            genome.add_groups(groups_file)

        return genome


    def stage_hub(
        self,
    ):

        with tempfile.TemporaryDirectory() as tmpdir:
            trackhub.upload.stage_hub(self._hub, staging=tmpdir)

            if self.description_url_path:
                description_basename = os.path.basename(self.description_url_path)
                with open(os.path.join(tmpdir, f"{self._hub.hub}.hub.txt"), "a") as hubtxt:
                    hubtxt.write("\n")
                    hubtxt.write(f"descriptionUrl {self.genome_name}/{description_basename}\n")

                shutil.copy(
                    self.description_url_path,
                    os.path.join(tmpdir, self.genome_name),
                )

            # Copy to the new location
            shutil.copytree(
                tmpdir,
                self.outdir,
                dirs_exist_ok=True,
                symlinks=False,
            )

            subprocess.run(["chmod", "-R", "2755", self.outdir])
        
    
    def to_pickle(self, path: str) -> None:
        """Save the hub to a pickle file"""
        with open(path, "wb") as f:
            pickle.dump(self, f)
            
            
    
   

        
    
    


import trackhub
import pandas as pd
from typing import Dict, Tuple
import seaborn as sns


def get_genome_file(genome_name: str, custom_genome: bool = False, **kwargs):

    if not custom_genome:
        genome = trackhub.Genome(genome_name)
        groups_file = None
    else:
        genome = trackhub.Assembly(
            genome=genome_name,
            twobit_file=kwargs["genome_twobit"],
            organism=kwargs["genome_organism"],
            defaultPos=kwargs.get("hub_default_position", "chr1:10000-20000"),
        )

        groups_file = trackhub.GroupsFile(
            [
                trackhub.GroupDefinition(
                    name=kwargs["hub_name"], priority=1, default_is_closed=False
                ),
            ]
        )

        genome.add_groups(groups_file)

    return genome


def make_track_palette(
    df_file_attributes: pd.DataFrame,
    palette: str,
    color_by: Tuple = None,
):

    if len(color_by) == 1:
        unique_samples = df_file_attributes[color_by[0]].unique()
    else:
        unique_samples = (
            df_file_attributes[color_by]
            .drop_duplicates()
            .astype(str)
            .apply("".join, axis=1)
        )
    colors = sns.color_palette(palette=palette, n_colors=len(unique_samples))
    color_mapping = dict(zip(unique_samples, colors))
    return color_mapping


import trackhub
import pandas as pd
from typing import Dict, Tuple
import re
import pathlib

from .track import get_track_subgroups

def get_subgroup_definitions(
    df_file_attributes: pd.DataFrame, grouping_columns: str = None
):

    columns_to_drop = ["fn", "path", "basename", "ext", "name"]

    if grouping_columns:
        columns_to_drop.append(grouping_columns)

    group_members = {
        col: df_file_attributes[col].unique()
        for col in df_file_attributes.drop(columns=columns_to_drop, errors="ignore")
    }

    subgroups = {
        grouping: trackhub.SubGroupDefinition(
            name=grouping,
            label=grouping.capitalize(),
            mapping={str(g).lower(): g for g in group},
        )
        for grouping, group in group_members.items()
    }

    return subgroups


def add_hub_group(container: trackhub.Track, hub: trackhub.Hub = None, custom_genome: bool = False):
    if custom_genome:
        if hub:
            container.add_params(group=hub.hub)
        else:
            raise ValueError("Custom genome specified but no hub provided")


def add_composite_tracks_to_container(
    container: trackhub.TrackDb,
    track_details: pd.DataFrame,
    subgroup_definitions: Dict[str, trackhub.SubGroupDefinition],
    color_by: list,
    color_mapping: Dict[str, Tuple[float, float, float]],
    track_suffix: str = "",
    custom_genome: bool = False,
    genome_twobit: str = "",  
    hub: trackhub.Hub = None,
):

    dimensions = dict(
        zip([f"dim{d}" for d in ["X", "Y", "A", "B", "C", "D"]], subgroup_definitions)
    )

    for track_type, df in track_details.groupby("ext"):

        track_types = {
            "bigwig": "Signal_Pileup",
            "bigbed": "Peaks_or_Regions_of_Interest",
        }


        # Make composite for each file type
        composite = trackhub.CompositeTrack(
            name=f"{track_types.get(str(track_type.lower()), 'Tracks')}",
            short_label=f"{track_type}",
            dimensions=" ".join([f"{k}={v}" for k, v in dimensions.items()]) if dimensions else None,
            sortOrder=" ".join([f"{k}=+" for k in subgroup_definitions]) if subgroup_definitions else None,
            tracktype=track_type,
            visibility="hide",
            dragAndDrop="subTracks",
            allButtonPair="off",
        )
        
        # Add group if custom genome
        add_hub_group(container=composite, hub=hub, custom_genome=custom_genome)
            

        # Add the subgroup definitions
        composite.add_subgroups(
            [definition for definition in subgroup_definitions.values()]
        )

        # Add tracks to composite
        for track_file in df.itertuples():
            

            path = pathlib.Path(track_file.path)
            track_name_base = trackhub.helpers.sanitize(track_file.name)
            track_groupings = get_track_subgroups(track_file, subgroup_definitions)

            track = trackhub.Track(
                name=f"{track_name_base}_{track_type}{'_' + track_suffix if track_suffix else ''}",
                shortLabel=" ".join(re.split(r"[.|_|\s+|-]", track_file.name)),
                longLabel=" ".join(re.split(r"[.|_|\s+|-]", track_file.name)),
                source=track_file.path,
                autoScale="on",
                tracktype=track_type,
                windowingFunction="mean",
                subgroups={
                    subgroup: str(getattr(track_file, subgroup)).lower()
                    for subgroup in subgroup_definitions
                },
                color=",".join(
                    [
                        str(int(x * 255))
                        for x in color_mapping[
                            "".join(getattr(track_file, cb) for cb in color_by)
                        ]
                    ]
                ),
            )

            # Add group to track if custom genome
            add_hub_group(container=track, hub=hub, custom_genome=custom_genome)

            # Append track to composite
            composite.add_subtrack(track)

        # Add to trackdb
        container.add_tracks(composite)


def add_overlay_track_to_container(
    track_name: str,
    container: trackhub.TrackDb,
    track_details: pd.DataFrame,
    color_by: list,
    color_mapping: Dict[str, Tuple[float, float, float]],
    custom_genome: bool = False,
    hub: trackhub.Hub = None,
):

    overlay = trackhub.AggregateTrack(
        aggregate="transparentOverlay",
        visibility="full",
        tracktype="bigWig",
        maxHeightPixels="8:80:50",
        showSubtrackColorOnUi="on",
        name=track_name,
    )

    # Need to add group if custom genome
    add_hub_group(container=overlay, hub=hub, custom_genome=custom_genome)

    for track_file in track_details.itertuples():

        track_name_base = trackhub.helpers.sanitize(track_file.name)

        track = trackhub.Track(
            name=f"{track_name_base}_{track_name}_overlay",
            shortLabel=" ".join(re.split(r"[.|_|\s+|-]", track_file.name)),
            longLabel=" ".join(re.split(r"[.|_|\s+|-]", track_file.name)),
            source=track_file.path,
            autoScale="on",
            tracktype="bigWig",
            windowingFunction="mean",
            color=",".join(
                [
                    str(int(x * 255))
                    for x in color_mapping[
                        "".join(getattr(track_file, cb) for cb in color_by)
                    ]
                ]
            ),
        )

        # Add group to track if custom genome
        add_hub_group(container=track, hub=hub, custom_genome=custom_genome)

        # Append track to composite
        overlay.add_subtrack(track)

        # Add to trackdb
        container.add_tracks(overlay)




def add_generic_tracks(
    parent: trackhub.BaseTrack,
    track_details: pd.DataFrame,
    color_by: list,
    color_mapping: Dict[str, Tuple[float, float, float]],
    custom_genome: bool = False,
    hub: trackhub.Hub = None,
):

    for track_file in track_details.itertuples():

        track = trackhub.Track(
            name=track_file.samplename,
            source=track_file.path,
            tracktype=track_file.ext.strip("."),
            autoScale="on",
            color=",".join(
                [
                    str(int(x * 255))
                    for x in color_mapping[
                        "".join(getattr(track_file, cb) for cb in color_by)
                    ]
                ]
            ),
        )

        add_hub_group(container=track, hub=hub, custom_genome=custom_genome)
        parent.add_tracks(track)

